import time
from flask import Blueprint, app, request, jsonify, json
from sqlalchemy import and_, select, join, text
from sqlalchemy import text
from datetime import datetime, timedelta
from app import db
from app.location_log.routes import add_location_log
from app.sensor_data.models import SensorData, Rawdata, Mapdata, Settings
from app.logs.models import AdvFrameType
from app.devices.models import Device
from app.gateways.models import Gateways
from app.device_types.models import DeviceType
from app.devices.distance import trilaterate
from app.utils import create_alert_for_notify_zero

sensor_data = Blueprint("sensor_data", __name__)
raw_data = Blueprint("raw_data", __name__)
map_data = Blueprint("map_data", __name__)
settings = Blueprint("settings", __name__)


@sensor_data.route("/api/v1/get-device-data", methods=["GET"])
def get_device_data():
    try:
        all_sensordata = SensorData.query.all()
        sensordata = []
        for sensor_data in all_sensordata:
            sensordata.append({
                'ga_mac': sensor_data.ga_mac,
                'dt_mac': sensor_data.dt_mac,
                'rssi_value': sensor_data.rssi_value,
                'measured_power': sensor_data.measured_power,
                'distance': sensor_data.distance
            })
        return jsonify(sensordata), 200
    except Exception as e:
        return jsonify({'error': str(e)}), 500

last_truncate_time = datetime.utcnow() - timedelta(hours=1)


@sensor_data.route("/api/v1/insert-device-data", methods=["POST"])
def insert_device_data():
    try:
        global last_truncate_time
        if datetime.utcnow() - last_truncate_time > timedelta(minutes=5):
            db.session.execute(text('TRUNCATE TABLE raw_data'))
            db.session.commit()
            last_truncate_time = datetime.utcnow()
        if datetime.utcnow() - last_truncate_time > timedelta(hours=24):
            db.session.execute(text('TRUNCATE TABLE map_data'))
            db.session.commit()
            last_truncate_time = datetime.utcnow()
        """Insert device data to database"""
        dataRows = request.get_json()
        rssi_values = []
        dt_mac = []
        dt_mac_rssi_pairs = []
        dt_mac_set = set()
        aset_tracker_devices = Device.query.with_entities(Device.device_mac_address).all()

        aset_tracker_dt_macs = [device.device_mac_address for device in aset_tracker_devices]

        gateways = Gateways.query.with_entities(Gateways.gateway_mac_address).all()

        gateway_ga_macs = [gateway.gateway_mac_address for gateway in gateways]

        settings = Settings.query.first()
        gatway_mac = None
        Single = 0
        Double = 0
        Long = 0
        temp_data = []
        temp_data1 = []
        temp_data2 = []
        if dataRows is None:
            return "No data received", 200
        else:
            if 'payload' in dataRows:
                dataRows = json.loads(dataRows['payload'])
            records_to_update = SensorData.query.filter(
                text("sensor_data.updated_at <= DATE_SUB(NOW(), INTERVAL 2 MINUTE)")).all()
            for record in records_to_update:
                record.distance = 0
                record.temp_distance = 0
                record.measured_power = None
            db.session.commit()
            n = 0
            for p in aset_tracker_dt_macs:
                rssi_values.append([])
            for row in dataRows:
                # print(row)
                for k, v in row.items():
                    if k == 'GatewayMAC' and v in gateway_ga_macs:
                        gatway_mac = v
                    if k == 'BLEMAC' and v in aset_tracker_dt_macs and gatway_mac is not None:
                        # print(v)
                        # print(row)
                        dt_index = aset_tracker_dt_macs.index(v)
                        if v not in dt_mac:
                            dt_mac.append(v)

                        for i, j in row.items():
                            if i == 'AdvFrameType':
                                call_action = j
                            if i == 'AlarmCount':
                                count = j

                                if call_action and count is not None:
                                    if call_action == 'Single':
                                        Single = count
                                        # print("Debug: Single count is", Single)
                                    elif call_action == 'Double':
                                        Double = count
                                        # print("Debug: Double count is", Double)
                                    elif call_action == 'Long':
                                        Long = count
                                        # print("Debug: Long count is", Long)
                                    try:
                                        device = Device.query.filter_by(device_mac_address=v).first()
                                        if not device:
                                            return
                                        device_id = device.device_id
                                        a_f_type = AdvFrameType()
                                        a_f_type.call_action = call_action
                                        a_f_type.count = count
                                        a_f_type.dt_mac = v
                                        a_f_type.device_id = device_id
                                        existing_record = AdvFrameType.query.filter_by(dt_mac=v,
                                                                                       call_action=call_action).first()
                                        if existing_record:
                                            if existing_record.count < count:
                                                existing_record.notify = 0
                                                existing_record.flag = 0
                                                existing_record.count = count
                                                db.session.commit()
                                        else:
                                            db.session.add(a_f_type)
                                            db.session.commit()
                                    except Exception as e:
                                        # print(f"Error: {e}")
                                        db.session.rollback()
                            if i == 'RSSI':
                                rssi_values[dt_index].append(j)
                                rssi_values[dt_index].sort(reverse=True)
                                middle_index = len(rssi_values[dt_index]) // 2
                                if len(rssi_values[dt_index]) % 2 == 0:  # if even number of elements
                                    middle_rssi_values = rssi_values[dt_index][middle_index - 1:middle_index + 1]
                                else:
                                    middle_rssi_values = rssi_values[dt_index][middle_index - 1:middle_index + 2]
                                mean_rssi = sum(middle_rssi_values) / len(middle_rssi_values)
                                # print(rssi_values[dt_index])
                                print(row)
                                if v not in dt_mac_set:
                                    dt_mac_set.add(v)
                                    dt_mac_rssi_pairs.append((v, mean_rssi))
                                    # lock_data = {}
            rec = len(dt_mac_rssi_pairs)

            for v, mean_rssi in dt_mac_rssi_pairs:
                raw_data = Rawdata()
                sensor_data = SensorData()
                map_data = Mapdata()
                raw_data.dt_mac = v
                raw_data.rssi_value = mean_rssi
                raw_data.ga_mac = gatway_mac
                sensor_data.ga_mac = gatway_mac
                sensor_data.dt_mac = v
                map_data.dt_mac = v
                sensor_data.rssi_value = mean_rssi
                sensor_data.cur_time = row['TimeStamp']
                db.session.add(raw_data)
                db.session.commit()
                if sensor_data.dt_mac and gatway_mac:
                    if sensor_data.measured_power is None and sensor_data.rssi_value is not None:
                        query = (
                            select(DeviceType.measured_power)
                            .select_from(join(Device, DeviceType, Device.device_type_id == DeviceType.device_type_id))
                            .filter(Device.device_mac_address == sensor_data.dt_mac)
                        )
                        result = db.session.execute(query).fetchone()
                        sensor_data.measured_power = result[0] if result else -59

                    if sensor_data.distance is None:
                        sensor_data.distance = 0
                    existing_data = SensorData.query.filter(
                        and_(SensorData.dt_mac == sensor_data.dt_mac),
                        (SensorData.ga_mac == gatway_mac)).first()
                    if existing_data:
                        existing_data.measured_power = sensor_data.measured_power

                        if sensor_data.rssi_value is not None:
                            if existing_data.measured_power is not None and existing_data.rssi_value is not None:
                                distance = 10 ** ((existing_data.measured_power - existing_data.rssi_value) / (
                                            10 * settings.N))

                                existing_data.distance = distance
                                existing_data.temp_distance = distance
                                existing_data.rssi_value = sensor_data.rssi_value
                                existing_data.cur_time = sensor_data.cur_time

                                result = db.session.query(SensorData.distance,
                                                          SensorData.temp_distance,
                                                          Device.x_coordinate,
                                                          Gateways.room_id,
                                                          Gateways.zone_id,
                                                          Device.y_coordinate,
                                                          Gateways.gateway_mac_address
                                                          ).join(Gateways,
                                                                 SensorData.ga_mac == Gateways.gateway_mac_address) \
                                    .join(Device, SensorData.dt_mac == Device.device_mac_address) \
                                    .filter(and_(SensorData.dt_mac == sensor_data.dt_mac, SensorData.distance != 0)) \
                                    .order_by(SensorData.updated_at.desc()) \
                                    .limit(3).all()
                                roomId = 0
                                zoneId = 0
                                siteId = 0
                                tempDistance = 0
                                distances = []
                                known_points = []
                                d1 = Device.query.filter_by(
                                    device_mac_address=sensor_data.dt_mac).first()
                                s1 = SensorData.query.filter_by(dt_mac=sensor_data.dt_mac).all()
                               
                                for res in s1:
                                     if res.temp_distance < tempDistance or tempDistance == 0:
                                         tempDistance = res.temp_distance

                                resultDistance = db.session.query(SensorData.distance,
                                                                  SensorData.temp_distance,
                                                                  Device.x_coordinate,
                                                                  Gateways.room_id,
                                                                  Gateways.site_id,
                                                                  Device.y_coordinate,
                                                                  Gateways.zone_id,
                                                                  Gateways.gateway_mac_address) \
                                    .join(Gateways, SensorData.ga_mac == Gateways.gateway_mac_address) \
                                    .join(Device, SensorData.dt_mac == Device.device_mac_address) \
                                    .filter(and_(SensorData.dt_mac == sensor_data.dt_mac),
                                            (SensorData.temp_distance != 0)) \
                                    .order_by(SensorData.updated_at.desc()) \
                                    .all()
                                for data in resultDistance:
                                    if data.temp_distance <= tempDistance or tempDistance == 0:
                                        if data.temp_distance > 0:
                                            tempDistance = data.temp_distance
                                            roomId = data.room_id
                                            zoneId = data.zone_id
                                            siteId = data.site_id
                                            d1.room_id = roomId
                                            d1.zone_id = zoneId
                                            # d1.site_id = siteId
                                db.session.commit()
                                add_location_log(d1.device_id, d1.device_mac_address, siteId, d1.device_name, roomId,
                                                 d1.site_id)

                                if len(result) == 3:
                                    all_positive = True
                                    for data in result:
                                        distances.append(data.distance)
                                        known_points.append([data.x_coordinate, data.y_coordinate])
                                        if data.distance is None:
                                            all_positive = False
                                            break
                                    if all_positive:
                                        coordinates = trilaterate(known_points, distances)
                                        if coordinates is not None:
                                            map_data.x_coordinate = coordinates[0]
                                            map_data.y_coordinate = coordinates[1]
                                            temp_data.append(v)
                                            temp_data1.append(distances)
                                            temp_data2.append(coordinates)
                                            db.session.add(map_data)
                                            db.session.commit()
                                            if d1:
                                                d1.room_id = roomId
                                                d1.zone_id = zoneId
                                                d1.site_id = siteId
                                                d1.x_coordinate = coordinates[0]
                                                d1.y_coordinate = coordinates[1]
                                                db.session.commit()
                                            else:
                                                print("Device not found.")
                                            for s2 in s1:
                                                s2.distance = 0

                                        dt_mac.append(v)
                                        db.session.commit()
                    else:
                        db.session.add(sensor_data)
                        db.session.commit()
                rec = rec - 1
            if (rec == 0):
                create_alert_for_notify_zero()
        db.session.commit()
        return "I'm healthy!!", 200
    except Exception as e:
        print(e.args)
        return f"Error: {e}", 501

@map_data.route("/api/v1/get/map-data", methods=["GET"])
def get_map_data():
    try:
        latest_mapdata = Mapdata.query.order_by(Mapdata.updated_at.desc()).limit(6).all()

        mapdata = []
        for map_data in latest_mapdata:
            time_only = map_data.updated_at.strftime("%H:%M:%S")
            mapdata.append({
                'dt_mac': map_data.dt_mac,
                'x_coordinate': map_data.x_coordinate,
                'y_coordinate': map_data.y_coordinate,
                'updated_at': time_only,
            })
            db.session.commit()
        return jsonify(mapdata), 200
    except Exception as e:
        return jsonify({'error': str(e)}), 50

@sensor_data.route("/api/v1/add/measured_power_device_data", methods=["POST"])
def measured_power_device_data():
    try:
        ga_mac = request.json.get('ga_mac')
        dt_mac = request.json.get('dt_mac')
        measured_power = request.json.get('measured_power')

        if ga_mac and dt_mac:
            existing_data = SensorData.query.filter_by(ga_mac=ga_mac, dt_mac=dt_mac).first()
            if existing_data:
                existing_data.measured_power = measured_power
                db.session.commit()
                return "Data updated successfully", 200
            else:
                sensor_data = SensorData(ga_mac=ga_mac, dt_mac=dt_mac, measured_power=measured_power)
                db.session.add(sensor_data)
                db.session.commit()
                return "New data added successfully", 200
        else:
            return "Error: Both ga_mac and dt_mac are required.", 400
    except Exception as e:
        return f"Error: {e}", 500

@settings.route("/api/v1/add/constant-value", methods=["POST"])
def constant_value():
    try:
        N = request.json.get('N')
        c = request.json.get('c')

        if not 2 <= N <= 4:
            return jsonify({"error": "N must be between 2 and 4"}), 400

        existing_settings = Settings.query.first()

        if existing_settings:
            if existing_settings.N != N or existing_settings.c != c:
                existing_settings.N = N
                existing_settings.c = c
                db.session.commit()
                return jsonify({"message": "Constant value updated successfully"}), 200
            else:
                return jsonify({"message": "Constant value already up to date"}), 200
        else:
            settings = Settings(N=N, c=c)
            db.session.add(settings)
            db.session.commit()
            return jsonify({"message": "Constant value created successfully"}), 201
    except Exception as e:
        return jsonify({"error": str(e)}), 500